/*
 *  This file is part of TopoS2, and has components from HealpiX software.
 *
 *  TopoS2 is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  TopoS2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Healpix_cxx; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *
 *  Copyright (C) Pratyush Pranav
 *  \author Pratysh Pranav <pratyuze@gmail.com>, <pratyush.pranav@ens-lyon.fr>
 */


#include <iostream>
#include <cstdio>
#include <string>
#include <iostream>
#include <fstream>
#include <cmath>
#include <cstdlib>
#include <sstream>
#include <cstddef>
#include <iomanip>
#include <limits>
#include <fitshandle.h>
#include <healpix_base.h>
#include <healpix_map.h>
#include "healpix_data_io.h"
#include "healpix_map_fitsio.h" 
#include "xcomplex.h"
#include "error_handling.h"
#include "alm.h"
#include "paramfile.h"
#include "alm_healpix_tools.h"
#include "alm_powspec_tools.h"
#include "levels_facilities.h"
#include "lsconstants.h"
#include "announce.h"
#include "geom_utils.h"

using namespace std ;
const double PI = 3.141592653589793238463;

namespace {

double alpha (const pointing &p1, const pointing &p2)
  {
  vec3 v1(p1), v2(p2);
  vec3 dir(v2-v1);
  return orientation (v2, dir) - orientation (v1,dir);
  }

template<typename T> void degrade_pol (const Healpix_Map<T> &q1,
  const Healpix_Map<T> &u1, Healpix_Map<T> &q2, Healpix_Map<T> &u2,
  bool pessimistic)
  {
  planck_assert(q1.conformable(u1) && q2.conformable(u2), "map mismatch");
  planck_assert(q2.Nside()<q1.Nside(),"this is no degrade");
  int fact = q1.Nside()/q2.Nside();
  planck_assert (q1.Nside()==q2.Nside()*fact,
    "the larger Nside must be a multiple of the smaller one");

  int minhits = pessimistic ? fact*fact : 1;
#pragma omp parallel
{
  int m, npix=q2.Npix();
#pragma omp for schedule (static)
  for (m=0; m<npix; ++m)
    {
    int x,y,f;
    q2.pix2xyf(m,x,y,f);
    int hits = 0;
    xcomplex<double> sum = 0;
    for (int j=fact*y; j<fact*(y+1); ++j)
      for (int i=fact*x; i<fact*(x+1); ++i)
        {
        int opix = q1.xyf2pix(i,j,f);
        if (!(approx<double>(q1[opix],Healpix_undef)
            ||approx<double>(u1[opix],Healpix_undef)))
          {
          ++hits;
          xcomplex<double> val(q1[opix],u1[opix]);
          double ang=alpha(q2.pix2ang(m),q1.pix2ang(opix));
          xcomplex<double> mul(cos(2*ang),sin(2*ang));
          sum += val*mul;
          }
        }
    q2[m] = T((hits<minhits) ? Healpix_undef : sum.real()/hits);
    u2[m] = T((hits<minhits) ? Healpix_undef : sum.imag()/hits);
    }
}
  }
 
}



int main(int argc, const char** argv) {
  if (argc != 2) {
    std::cout<< "Usage: ./fits_RDS_mask_RDS_nrm param_file"<< std::endl;
    return -1;
  }


   
  paramfile params (getParamsFromCmdline(argc,argv));
  params.setVerbosity(false);
  int nside = params.template find<int>("nside");
  string infile = params.template find<string>("infile");
  string maskfile = params.template find<string>("maskfile");
  string outfile = params.template find<string>("outfile");
  float level = params.template find<float>("level");
  bool polarisation = params.template find<bool>("polarisation",false);
  bool pessimistic = params.template find<bool>("pessimistic",false);
  int nlmax = params.template find<int>("nlmax");
  int num_iter = params.template find<int>("iter_order",0);
  double fwhm_map = arcmin2rad*params.template find<double>("fwhm_map");
  double fwhm_mask = arcmin2rad*params.template find<double>("fwhm_mask");
  bool dp = params.find<bool> ("double_precision",true);


  fitshandle fitshandle;
  vec3 point;
  pointing coord;
  //template<typename T>
  //Healpix_Map< T > map;

  //Healpix_map map;
  // read a HEALPix map from a file -----------------------------
  std::cerr<<"Reading healpix map and mask...";
  Healpix_Map<double> map, mask_map ;
  read_Healpix_map_from_fits(infile, map,1,2);
  read_Healpix_map_from_fits(maskfile, mask_map,1,2);

  std::cerr<<"done\ndegrading data...";
  Healpix_Map<double> outmap (nside,map.Scheme(),SET_NSIDE);
  outmap.Import(map,pessimistic);
  //write_Healpix_map_to_fits (outfile,outmap,planckType<T>());

  std::cerr<<"done\ndegrading mask...";

  Healpix_Map<double> degraded_mask (nside,mask_map.Scheme(),SET_NSIDE);
  degraded_mask.Import(mask_map,pessimistic);
 
  if (outmap.Scheme()==NEST) outmap.swap_scheme();
  if (degraded_mask.Scheme()==NEST) degraded_mask.swap_scheme();
  
  std::cerr<<"done\nsmoothing data...";
  tsize nmod = outmap.replaceUndefWith0();
    if (nmod!=0)
      cout << "WARNING: replaced " << nmod <<
              " undefined map pixels with a value of 0" << endl;

  double avg=outmap.average();
  outmap.Add(-avg);

  arr<double> weight;
  get_ring_weights (params,outmap.Nside(),weight);

  Alm<xcomplex<double> > alm(nlmax,nlmax);
  if (outmap.Scheme()==NEST) map.swap_scheme();

  map2alm_iter(outmap,alm,num_iter,weight);
  smoothWithGauss (alm, fwhm_map);
  alm2map(alm,outmap);

  outmap.Add(avg);

  std::cerr<<"done\nsmoothing mask...";
  nmod = degraded_mask.replaceUndefWith0();
    if (nmod!=0)
      cout << "WARNING: replaced " << nmod <<
              " undefined map pixels with a value of 0" << endl;

    //double avg=degraded_mask.average();
  degraded_mask.Add(-avg);

  //arr<double> weight;
  get_ring_weights (params,degraded_mask.Nside(),weight);

  //Alm<xcomplex<double> > alm(nlmax,nlmax);
  if (degraded_mask.Scheme()==NEST) degraded_mask.swap_scheme();

  map2alm_iter(degraded_mask,alm,num_iter,weight);
  smoothWithGauss (alm, fwhm_mask);
  alm2map(alm,degraded_mask);

  degraded_mask.Add(avg);

  //if (map.Scheme()==NEST) map.swap_scheme();
  if (outmap.Scheme()==NEST) outmap.swap_scheme();
  if (degraded_mask.Scheme()==NEST) degraded_mask.swap_scheme();

  std::cerr<<"done\n setting cutoff at "<<level<<"and fixing smoothed mask...";
 
  for (long i = 0; i<degraded_mask.Npix(); i++)
   {
    if(degraded_mask[i]>=level) degraded_mask[i]=1;
    else degraded_mask[i]=0;
  }

  std::cerr<<"done\n Normalizing temperature and masking...";

  double sdev = 0;
  long count = 0;
  avg = 0;

  for (long i = 0; i<outmap.Npix(); i++)
    {
      coord = outmap.pix2ang(i);
      //if(coord.theta <= 45*PI/180 || coord.theta >= 135*PI/180)
      if(degraded_mask[i]>0)
	{
	  avg+=outmap[i];
	  count+=1;
	}
    }
  avg/=count;


  for (long i = 0; i<outmap.Npix(); i++)
    {
      coord = outmap.pix2ang(i);
      //if(coord.theta <= 45*PI/180 || coord.theta >= 135*PI/180)
      if(degraded_mask[i]>0)
	sdev+=pow(outmap[i]-avg,2);
    }
    sdev=sqrt(sdev/(count-1));

  for (long i = 0; i<outmap.Npix(); i++)
    {
      coord = outmap.pix2ang(i);
      //if(coord.theta <= 45*PI/180 || coord.theta >= 135*PI/180)
      if(degraded_mask[i]>0)
	outmap[i]=(outmap[i]-avg)/sdev;
      else
	outmap[i]=21;
    }

  //std::cerr<<"done\nwriting fits file...";

  //write_Healpix_map_to_fits (outfile,outmap,planckType<float>());

  std:cerr<<"done\nwriting streaming output\n";

  //cout<<map.Npix()<<"\n";
  
  for (long i = 0; i<outmap.Npix(); i++)
    {
      point = outmap.pix2vec(i);
      cout<<point.x<<"\t"<<point.y<<"\t"<<point.z<<"\t";
      cout<<std::setprecision (std::numeric_limits<double>::digits10 + 1)<<outmap[i]<<"\n";
    }
  //fitshandle.open(argv[1]);
  //std::cout<<"No of HDU's :"<<fitshandle.num_hdus();
  //handle=fitshandle.open(argv[1]);
  //handle.close();

  return 0;
}
