#ifndef LOWER_STAR_H
#define LOWER_STAR_H

#include "octree.h"
#include "Cube_vertex.h"
#include "Converter.h"
#include "Simplex.h"

#include <vector>


template<typename InputIterator,typename OutputIterator>
  inline OutputIterator construct_simplex(InputIterator node_begin, 
					  InputIterator node_end,
					  long id, 
					  bool is_lower,
					  OutputIterator out) {

  bool id_exists=false;
  for(InputIterator it=node_begin;it!=node_end;it++) {
    if(! (*it)) {
      return out;
    }
    if((*it)->id==id) {
      id_exists=true;
    }
  }
  if(! id_exists) {
    return out;
  }
  Simplex new_simplex;
  for(InputIterator it=node_begin;it!=node_end;it++) {
    new_simplex.push_back((*it)->id);
  }
  std::sort(new_simplex.begin(),new_simplex.end());
  for(int i=0;i<(int)new_simplex.size()-1;i++) {
    if(new_simplex[i]==new_simplex[i+1]) {
      return out;
    }
  }
  if(is_lower && new_simplex[new_simplex.size()-1]!=id) {
    return out;
  }
  if(!is_lower && new_simplex[0]!=id) {
    return out;
  }
  *out++=new_simplex;
  return out;
}

template<int D>
struct subset_iterator {
  std::vector<int> entries;
  int length;
  int curr;
  long end_val;
  long offset;


  template<typename InputIterator>
  subset_iterator(int init,InputIterator begin, InputIterator end) {
    curr=init;
    std::copy(begin,end,std::back_inserter(entries));
    length=(int)entries.size();
    end_val = (1 << length);
    InputIterator it=begin;
    offset=0;
    for(int i=0;i<D;i++) {
      if(it==end || i!=*it) {
	offset+= (1 << i);
      } else {
	it++;
      }
    }   
  }

  void next() {
    curr++;
  }

  bool done() {
    return curr>=end_val;
  }

  std::pair<int, std::vector<int> >
    operator* () {
    long vertex=offset;
    std::vector<int> values;
    for(int i=0;i<length;i++) {
      if(!(curr & (1<<i))) {
	values.push_back(entries[i]);
      } else {
	vertex+= (1 << entries[i]);
      }
    }
    return std::make_pair(vertex,values);    
  }
  
  };  

template<int D, typename OutputIterator>
  void freudenthal_intern(const std::vector<int>& entries,int d,OutputIterator simplices,bool toplevel) {
  subset_iterator<D> si(toplevel ? 0 : 1,entries.begin(),entries.end());
  while(! si.done()) {
    std::pair<int,std::vector<int> > curr = *si;
    if(d>0) {
      std::vector<Simplex> simplices_d_minus_1;
      freudenthal_intern<D>(curr.second,d-1,std::back_inserter(simplices_d_minus_1),false);
      for(int i=0; i < (int)simplices_d_minus_1.size();i++) {
	simplices_d_minus_1[i].push_back(curr.first);
      }
      std::copy(simplices_d_minus_1.begin(),simplices_d_minus_1.end(),simplices);
    } else {
      Simplex new_simpl;
      new_simpl.push_back(curr.first);
      *simplices++=new_simpl;
    }
    si.next();
  }
}


template<int D,typename OutputIterator> 
  OutputIterator freudenthal_triangulation
  (OutputIterator simplices) {

  std::vector<int> entries;
  for(int i=0;i<D;i++) {
    entries.push_back(i);
  }
  std::vector<Simplex> rev_simplices;
  for(int d=0;d<=D;d++) {
    // Find d-simplices
    freudenthal_intern<D>(entries,d,std::back_inserter(rev_simplices),true);
  }
  for(int i=0;i<(int)rev_simplices.size();i++) {
    std::reverse(rev_simplices[i].begin(),rev_simplices[i].end());
  }
  std::copy(rev_simplices.begin(),rev_simplices.end(),simplices);
  return simplices;
}

template<int D,typename InputIterator,typename OutputIterator> void simplices_for_vertex
    (Cube_vertex<D>& v,
     long id,
     bool is_lower,
     Ocnode<D>* vertex_at_infinity,
     InputIterator freudenthal_begin,
     InputIterator freudenthal_end,
     OutputIterator simplices) {

  find_all_adjacent_terminals(v);
  Ocnode<D>* c[Ocnode<D>::no_children];
  for(int i=0;i<Ocnode<D>::no_children;i++) {
    c[i]=v.adj_terminals[i];
    if(!c[i]) {
      c[i]=vertex_at_infinity;
    }
  }
  for(InputIterator it=freudenthal_begin;it!=freudenthal_end;it++) {
    if(it->size()==1) {
      continue;
    }
    std::vector<Ocnode<D>*> curr_nodes;
    for(int i=0;i<(int)it->size();i++) {
      curr_nodes.push_back(c[(*it)[i]]);
    }
    construct_simplex(curr_nodes.begin(),curr_nodes.end(),id,is_lower,simplices);
  }
}  

template<int D,typename OutputIterator>
  OutputIterator get_patterns_for_neighbors(OutputIterator patterns) {

#if 0  
  long no = (long)pow(3,D);
  for(long i=0;i<no-1;i++) {
    long offset=0;
    std::vector<int> free_vars;
    std::vector<int> path_to_neighbor;
    long k=i;
    for(int j=0;j<D;j++) {
      int curr_info = k%3;
      if(curr_info==1) {
	path_to_neighbor.push_back((j<<1)+1);
      } else if(curr_info==0) {
	offset+=(1<<j); 
	path_to_neighbor.push_back(j<<1);
      } else if(curr_info==2) {
	free_vars.push_back(j);
      }
      k=k/3;
    }
    std::vector<int> curr_pattern;
    int no_free=free_vars.size();
    
    if(no_free==0) {
      continue;
    }
    
    for(int j=0;j<(1<<no_free);j++) {
      int curr_index=offset;
      for(int k=0;k<no_free;k++) {
	if(j & (1<<k)) {
	  curr_index+=(1<<free_vars[k]);
	}
      }
      curr_pattern.push_back(curr_index);
    }
    *patterns++=std::make_pair(path_to_neighbor,curr_pattern);
  }
  return patterns;
#else
  for(long i=0;i<Ocnode<D>::no_children;i++) {
    long offset=0;
    std::vector<int> free_vars;
    std::vector<int> path_to_neighbor;
    for(int j=0;j<D;j++) {
      if(i & (1 << j)) {
	offset+=(1<<j); 
	path_to_neighbor.push_back(j<<1);
      } else {
	free_vars.push_back(j);
      }
    }
    if(path_to_neighbor.empty()) {
      continue;
    }
    std::vector<int> curr_pattern;
    int no_free=free_vars.size();
    
    for(int j=0;j<(1<<no_free);j++) {
      int curr_index=offset;
      for(int k=0;k<no_free;k++) {
	if(j & (1<<k)) {
	  curr_index+=(1<<free_vars[k]);
	}
      }
      curr_pattern.push_back(curr_index);
    }
    *patterns++=std::make_pair(path_to_neighbor,curr_pattern);

    curr_pattern.clear();
    for(int j=0;j<(int)path_to_neighbor.size();j++) {
      path_to_neighbor[j]++;
    }

    for(int j=0;j<(1<<no_free);j++) {
      int curr_index=0;
      for(int k=0;k<no_free;k++) {
	if(j & (1<<k)) {
	  curr_index+=(1<<free_vars[k]);
	}
      }
      curr_pattern.push_back(curr_index);
    }
    *patterns++=std::make_pair(path_to_neighbor,curr_pattern);

  }
  return patterns;
#endif

}
  
template<int D,typename InputIterator,typename OutputIterator> 
  OutputIterator check_neighbor_for_star
    (Ocnode<D> *root,
     std::vector<int> pattern,
     long id,
     bool is_lower,
     Ocnode<D>* vertex_at_infinity,
     InputIterator freudenthal_begin,
     InputIterator freudenthal_end,
     OutputIterator simplices) {
  if(! root->is_interior_after_prune) {
    return simplices;
  }
  std::queue<Ocnode<D>*> nodes;
  std::vector<Cube_vertex<D> > vertices_on_boundary;
  nodes.push(root);
  while(! nodes.empty()) {
    Ocnode<D>* curr=nodes.front();
    nodes.pop();
    for(int i=0;i<(int)pattern.size();i++) {
      if(curr->is_terminal) {
	vertices_on_boundary.push_back(Cube_vertex<D>(curr,pattern[i]));
      } else {
	assert(curr->is_interior_after_prune);
	if(curr->child[pattern[i]]) {
	  nodes.push(curr->child[pattern[i]]);
	}
      }
      
    }
  }
  
  std::sort(vertices_on_boundary.begin(),
	    vertices_on_boundary.end(),
	    Cube_vertex_compare<D>());
  typename std::vector<Cube_vertex<D> >::iterator uniq_end =
    std::unique(vertices_on_boundary.begin(),
		vertices_on_boundary.end());
  
  for(typename std::vector<Cube_vertex<D> >::iterator it
	=vertices_on_boundary.begin();
      it!=uniq_end;
      it++) {
    simplices_for_vertex(*it,id,is_lower,
			 vertex_at_infinity,
			 freudenthal_begin,
			 freudenthal_end,
			 simplices);
  }
  return simplices;
}

template<int D,typename InputIterator1, typename InputIterator2, typename OutputIterator>
   OutputIterator star(Ocnode<D> *node, Ocnode<D> *root, Ocnode<D> *vertex_at_infinity,
		       bool is_lower, 
		       InputIterator1 freudenthal_begin,
		       InputIterator1 freudenthal_end,
		       InputIterator2 neighbor_patterns_begin,
		       InputIterator2 neighbor_patterns_end,
		       OutputIterator simplices) {

  std::vector<Simplex> star_simplices;
  
  long id=node->id;

  if(node!=vertex_at_infinity) {

    for(int i=0;i<Ocnode<D>::no_children;i++) {
      Cube_vertex<D> v(node,i);
      simplices_for_vertex(v,id,is_lower,vertex_at_infinity,
			   freudenthal_begin,freudenthal_end,
			   std::back_inserter(star_simplices));
    }
    
    for(InputIterator2 it=neighbor_patterns_begin;it!=neighbor_patterns_end;it++) {
      std::vector<int> path=it->first;
      std::vector<int> pattern=it->second;
      Ocnode<D>* neighbor=node;
      for(int j=0; j<(int)path.size();j++) {
	neighbor=neighbor->neighbor[path[j]];
	if(! neighbor) {
	  break;
	}
      }
      if(! neighbor) {
	continue;
      }
      check_neighbor_for_star
	(neighbor,pattern,id,is_lower,vertex_at_infinity,
	 freudenthal_begin,freudenthal_end,std::back_inserter(star_simplices));
    }

  } else {
    /* TODO Fix
    // do something else: just run through all terminal node
    std::queue<Ocnode<D>*> nodes;
    nodes.push(root);
    while(!nodes.empty()) {
      Ocnode<D>* curr=nodes.front();
      nodes.pop();
      if(!curr->is_terminal) {
	assert(curr->is_interior_after_prune);
	for(int i=0;i<Ocnode<D>::no_children;i++) {
	  if(curr->child[i]) {
	    nodes.push(curr->child[i]);
	  }
	}
      } else {
	if(!curr->front) {
	  short indices[4] = { 0,2,4,6 };
	  for(int i=0;i<4;i++) {
	    Cube_vertex v(curr,indices[i]);
	    simplices_for_vertex(v,id,is_lower,vertex_at_infinity,
				 std::inserter(edges,edges.begin()),
				 std::inserter(faces,faces.begin()),
				 std::inserter(tetrahedra,tetrahedra.begin()) );
	  }
	}
	if(!curr->back) {
	  short indices[4] = { 1,3,5,7 };
	  for(int i=0;i<4;i++) {
	    Cube_vertex v(curr,indices[i]);
	    simplices_for_vertex(v,id,is_lower,vertex_at_infinity,
				 std::inserter(edges,edges.begin()),
				 std::inserter(faces,faces.begin()),
				 std::inserter(tetrahedra,tetrahedra.begin()) );
	  }
	}
	if(!curr->above) {
	  short indices[4] = { 2,3,6,7 };
	  for(int i=0;i<4;i++) {
	    Cube_vertex v(curr,indices[i]);
	    simplices_for_vertex(v,id,is_lower,vertex_at_infinity,
				 std::inserter(edges,edges.begin()),
				 std::inserter(faces,faces.begin()),
				 std::inserter(tetrahedra,tetrahedra.begin()) );
	  }
	}
	if(!curr->below) {
	  short indices[4] = { 0,1,4,5 };
	  for(int i=0;i<4;i++) {
	    Cube_vertex v(curr,indices[i]);
	    simplices_for_vertex(v,id,is_lower,vertex_at_infinity,
				 std::inserter(edges,edges.begin()),
				 std::inserter(faces,faces.begin()),
				 std::inserter(tetrahedra,tetrahedra.begin()) );
	  }
	}
	if(!curr->left) {
	  short indices[4] = { 0,1,2,3 };
	  for(int i=0;i<4;i++) {
	    Cube_vertex v(curr,indices[i]);
	    simplices_for_vertex(v,id,is_lower,vertex_at_infinity,
				 std::inserter(edges,edges.begin()),
				 std::inserter(faces,faces.begin()),
				 std::inserter(tetrahedra,tetrahedra.begin()) );
	  }
	}
	if(!curr->right) {
	  short indices[4] = { 4,5,6,7 };
	  for(int i=0;i<4;i++) {
	    Cube_vertex v(curr,indices[i]);
	    simplices_for_vertex(v,id,is_lower,vertex_at_infinity,
				 std::inserter(edges,edges.begin()),
				 std::inserter(faces,faces.begin()),
				 std::inserter(tetrahedra,tetrahedra.begin()) );
	  }
	}
      }
    }
    */
  }


  Simplex simplex_for_vertex;
  simplex_for_vertex.push_back(node->id);
  star_simplices.push_back(simplex_for_vertex);
  std::sort(star_simplices.begin(),star_simplices.end(),Simplex_sort());
  std::vector<Simplex>::iterator uniq_end =
    std::unique(star_simplices.begin(),star_simplices.end());
  std::copy(star_simplices.begin(),uniq_end,simplices);
  return simplices;
}



#endif
