#ifndef OCTREE_H
#define OCTREE_H 1

#include "flags.h"
#include "Converter.h"
#include "Simplex.h"

#include <math.h>
#include <fstream>
#include <string>
#include <sstream>
#include <algorithm>
#include <cassert>
#include <vector>
#include <queue>
#include <limits>

template<int D> 
struct Ocnode {
  
  typedef Ocnode<D> Self;

  long coordinates[2*D];
  long id;
  Self* parent;
  Index_type index;
  short level;
  bool is_terminal,is_leaf;
  bool is_interior_after_prune;
  static const Index_type no_children = (1 << D);
  Self* child[no_children];
  Self* neighbor[2*D];
  float min,max, value;
#if USES_COMPACTIFIED_SPHERE
  bool is_compactification;
#endif
  float variation;
  Ocnode() {}

#if USES_COMPACTIFIED_SPHERE
  Ocnode(bool is_compactification) : is_compactification(is_compactification) {
    if(is_compactification) {
      is_terminal=true;
      is_interior_after_prune=false;
      for(int i=0;i<2*D;i++) {
	neighbor[i]=this;
      }
      min=max=std::numeric_limits<float>::has_infinity
	? std::numeric_limits<float>::infinity()
	: std::numeric_limits<float>::max();
      for(int i=0;i<no_children;i++) {
	child[i]=NULL;
      }
    }
  }
#endif
      
    Ocnode(int* coords,
	   Self* parent,
	   Index_type index, short level, float* P, const Converter<D>& cv) 
    : id(-1),parent(parent),index(index),
      level(level),
      is_terminal(false),
      is_interior_after_prune(false)
      //no_children((int)pow(2.,D))
#if USES_COMPACTIFIED_SPHERE    
    , is_compactification(false)
#endif
    {
      for(int i=0;i<no_children;i++) {
	child[i]=NULL;
      }
      for(int j=0;j<2*D;j++) {
	coordinates[j]=coords[j];
      }
      is_leaf=true;
      for(int i=0;i<D;i++) {
	if(coordinates[2*i]!=coordinates[2*i+1]) {
	  is_leaf=false;
	  break;
	}
      }
      for(int i=0;i<2*D;i++) {
	neighbor[i]=NULL;
      }
      
      if(!is_leaf) {
	min=max=0.;
      } else {
	int cv_coords[D];
	for(int i=0;i<D;i++) {
	  cv_coords[i]=coords[2*i];
	  assert(cv_coords[i]==coords[2*i+1]);
	}
	min = P[cv(cv_coords)];
	max = P[cv(cv_coords)]; 
	value = min;
      }
    }

  /*  
  virtual ~Ocnode() {
    delete[] child;
  }
  */
  

};

template <int D>
std::ostream& operator<<(std::ostream& str, Ocnode<D>* org_node) {
  Ocnode<D>* node=org_node;
  if(!node) {
    str << "NULL";
    return str;
  }
#if USES_COMPACTIFIED_SPHERE
  if(node->is_compactification) {
    str << "inf";
    return str;
  }
#endif
  std::vector<long> indices;
  while(node->index!=-1) {
    indices.push_back(node->index);
    node=node->parent;
  }
  for(unsigned int i=0;i<indices.size();i++) {
    str << indices[indices.size()-i-1] << " ";
  }
  //if(node->id>=0) {
    str << "(id=" << org_node->id << ", " << "minmax=" << org_node->min << ", " << org_node->max << " )";
    //}
  return str;
}

template<int D> 
void clear_all_nodes(Ocnode<D>* node) {
  node->id=-1;
  node->is_terminal=false;
  node->is_interior_after_prune=false;
  for(int i=0;i<node->no_children;i++) {
    if(node->child[i]) {
      clear_all_nodes(node->child[i]);
    }
  }
}

template<int D>
struct Tree_builder {
  
  int bounds[D];
  
  int grid;

  long node_count;

  Converter<D> cv;

  Ocnode<D>* nodes;

  Tree_builder(int* bounds,Ocnode<D>* nodes) 
    : cv(bounds), nodes(nodes) {
    int M = 0;
    for(int i=0;i<D;i++) {
      this->bounds[i]=bounds[i];
      M = std::max(M,bounds[i]);
    }
    grid = 1;
    while(grid<M) {
      grid*=2;
    }
    node_count=0;
  }

  Ocnode<D>* operator() (float* P) {
    int initals[2*D];
    for(int i=0;i<D;i++) {
      initals[2*i]=0;
      initals[2*i+1]=grid-1;
    }
    build_up(initals,0,-1,0,P);
    return &nodes[0];
  }

  bool build_up(int* coords,Ocnode<D>* parent,
		short index,short level,float* P) {
    bool is_padded=false;
    for(int i=0;i<D;i++) {
      if(coords[2*i]>=bounds[i]) {
	is_padded=true;
	break;
      }
    }
    if(is_padded) {
      return false;
    } else {
      nodes[node_count++]
	= Ocnode<D>(coords,parent,index,level,P,cv);
    }

    Ocnode<D>* root=&nodes[node_count-1];

    if(root->is_leaf) {
      return true;
    }
    for(Index_type i=0;i<Ocnode<D>::no_children;i++) {
      int new_coords[2*D];
      for(int j=0;j<D;j++) {
	Index_type curr = (i & (1 << j));
	int H = (coords[2*j]+coords[2*j+1])/2;
	if(curr) {
	  new_coords[2*j] = H+1;
	  new_coords[2*j+1] = coords[2*j+1]; 
	} else {
	  new_coords[2*j] = coords[2*j];
	  new_coords[2*j+1] = H; 
	}
      }
      int curr_count=node_count;
      if(build_up(new_coords,root,i,level+1,P)) {
	root->child[i]=&nodes[curr_count];
      }
    }

    return true;
  }

};

template<int D>
void print_tree(Ocnode<D>* node, std::string padding) {
  std::cerr << padding << node << std::endl;
  for(int i=0;i<Ocnode<D>::no_children;i++) {
    if(node->child[i] && !node->is_terminal) {
      print_tree(node->child[i],padding+"  ");
    }
  }
}

template<int D>
void set_neighbor(Ocnode<D>* node, int k) {
  Index_type index = node->index;
  if(index==-1) {
    // the root
#if EMBED_ON_TORUS
    node->neighbor[2*k] = node;
    node->neighbor[2*k+1] = node;
#endif
    return;
  }
  assert(node->parent != NULL);
  Ocnode<D>* front_partner=NULL;
  if(index & (1<< k)) {
    front_partner = node->parent->child[index^(1 << k)];
  } else {
    Ocnode<D>* par_front_partner = node->parent->neighbor[2*k];
    if(par_front_partner) {
      front_partner=par_front_partner->child[index^(1 << k)];
    }
  }
  node->neighbor[2*k]=front_partner;
  if(front_partner) {
    front_partner->neighbor[2*k+1]=node;
  }
}

template<int D>
void set_all_neighbors(Ocnode<D>* node) {
  for(int i=0;i<D;i++) {
    set_neighbor(node,i);
#if EMBED_ON_TORUS
    assert(node->neighbor[2*i]!=NULL);
#endif
  }
  for(int i=0;i<Ocnode<D>::no_children;i++) {
    if(node->child[i]) {
      set_all_neighbors(node->child[i]);
    }
  }
}

template<int D,typename OutputIterator> OutputIterator assign_ids(Ocnode<D>* root,OutputIterator terminals_id_sorted) {
  std::vector<std::pair<float, Ocnode<D>*> > terminal_vec;
  std::queue<Ocnode<D>*> nodes;
  nodes.push(root);
  while(! nodes.empty()) {
    Ocnode<D>* curr=nodes.front();
    nodes.pop();
    if(curr->is_terminal) {
      terminal_vec.push_back(std::make_pair(curr->value,curr));
    } else {
      assert(curr->is_interior_after_prune);
      for(int i=0;i<Ocnode<D>::no_children;i++) {
	if(curr->child[i]) {
	  nodes.push(curr->child[i]);
	}
      }
    }
  }
  std::sort(terminal_vec.begin(),terminal_vec.end());
  long id_count=0;
  for(int i=0;i<(int)terminal_vec.size();i++) {
    terminal_vec[i].second->id=id_count++;
    *terminals_id_sorted++=terminal_vec[i];
  }
  return terminals_id_sorted;
}

template<int D>
void set_variation(Ocnode<D>* node) {
  if(! node->is_leaf) {
    float min = std::numeric_limits<float>::has_infinity
      ? std::numeric_limits<float>::infinity()
      : std::numeric_limits<float>::max();
    float max=-min;
#if SUBDIVISION_RULE==ABSOLUTE_VALUE_THRESHOLD
    float abs_min=min;
#endif
    for(int i=0;i<Ocnode<D>::no_children;i++) {
      if(node->child[i]) {
	set_variation(node->child[i]);
	min = (min < node->child[i]->min) ? min : node->child[i]->min;
	max = (max > node->child[i]->max) ? max : node->child[i]->max;
#if SUBDIVISION_RULE==ABSOLUTE_VALUE_THRESHOLD
	if(min<0 && max>0) {
	  abs_min=0.;
	} else {
	  abs_min = std::min(abs(min), abs(max));
	}
#endif
      }
    }
    node->min=min;
    node->max=max;
    node->value=(min+max)/2;


#if SUBDIVISION_RULE==ABSOLUTE_VALUE_THRESHOLD
    if(max==min) {
      node->variation=0; // no reason to subdivide
    } else {
      node->variation = (max-min) / abs_min;
    }
#endif
#if SUBDIVISION_RULE==GLOBAL_VALUE_THRESHOLD
    node->variation=max-min;
#endif
  }
}


template<int D>
bool _violates_balancing(Ocnode<D>* node, const std::vector<int>& index) {
  if(! node) {
    return false;
  }
  if(! node->is_interior_after_prune) {
    return false;
  }
  for(int i=0;i<(int)index.size();i++) {
    assert(node->child[index[i]]);
    if(node->child[index[i]]->is_interior_after_prune) {
      return true;
    }
  }
  return false;
}

template<int D, typename InputIterator>
  bool balance(Ocnode<D>* root, 
	       InputIterator pattern_begin, 
	       InputIterator pattern_end) {
  std::queue<Ocnode<D>*> nodes;
  nodes.push(root);
  bool was_balanced=true;
  while(! nodes.empty()) {
    Ocnode<D>* curr=nodes.front();
    nodes.pop();
    if(curr->is_interior_after_prune) {
      for(int i=0;i<Ocnode<D>::no_children;i++) {
	if(curr->child[i]) {
	  nodes.push(curr->child[i]);
	}
      }
    } else {
      assert(curr->is_terminal);
      for(InputIterator it=pattern_begin;it!=pattern_end;it++) {
	std::vector<int> path_to_neighbor=it->first;
	Ocnode<D>* neighbor=curr;
	for(int i=0; i<(int)path_to_neighbor.size();i++) {
	  neighbor=neighbor->neighbor[path_to_neighbor[i]];
	  if(! neighbor) {
	    break;
	  }
	}
	if(! neighbor) {
	  continue;
	}
	if(_violates_balancing(neighbor,it->second)) {
	  was_balanced=false;
	  curr->is_terminal=false;
	  curr->is_interior_after_prune=true;
	  for(int i=0;i<Ocnode<D>::no_children;i++) {
	    if(curr->child[i]) {
	      curr->child[i]->is_terminal=true;
	      nodes.push(curr->child[i]);
	    }
	  }
	}

      }
    }
  }
  if(! was_balanced) {
    balance(root,pattern_begin,pattern_end);
  }
  return !was_balanced;
}

template<int D> bool is_boundary(Ocnode<D>* node) {
  for(int i=0;i<2*D;i++) {
    if(!node->neighbor[i]) {
      return true;
    }
  }
  return false;
}

template<int D>
void subdivide_boundary(Ocnode<D>* root) {
  std::queue<Ocnode<D>*> nodes;
  nodes.push(root);
  while(! nodes.empty()) {
    Ocnode<D>* curr=nodes.front();
    nodes.pop();
    if(curr->is_interior_after_prune) {
      for(int i=0;i<Ocnode<D>::no_children;i++) {
	if(curr->child[i]) {
	  nodes.push(curr->child[i]);
	}
      }
    } else {  
      assert(curr->is_terminal);
      if(is_boundary(curr) && !curr->is_leaf) {
	curr->is_terminal=false;
	curr->is_interior_after_prune=true;
	for(int i=0;i<Ocnode<D>::no_children;i++) {
	  if(curr->child[i]) {
	    curr->child[i]->is_terminal=true;
	    nodes.push(curr->child[i]);
	  }
	}
      }
    }
  }
}

#endif


