/*
 *  This file is part of TopoS2.
 *
 *  TopoS2 is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  TopoS2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with TopoS2; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *
 *  Copyright (C) Pratyush Pranav, based on PHAT and Dionysus library, and contributions from Michael Kerber 
 *  \author Pratysh Pranav <pratyuze@gmail.com>, <pratyush.pranav@ens-lyon.fr>
 */




#ifndef USE_DIONYSUS_SIMPLEX
#define USE_DIONYSUS_SIMPLEX 0
#endif

#if USE_DIONYSUS_SIMPLEX
#include <topology/simplex.h>
#include <topology/cohomology-persistence.h>

#include <utilities/log.h>
#include <utilities/indirect.h>

#endif

#include <queue>

#include <vector>
#include <map>
#include <iostream>
#include <set>
#include <list>
#include <algorithm>
#include <cassert>
#include <sstream>
#include <limits>
#include <fstream>

//#include <boost/tuple/tuple.hpp>

#include <boost/timer.hpp>

#include <phat/representations/bit_tree_pivot_column.h>
#include <phat/algorithms/chunk_reduction.h>

#include<ctime>
#include <limits>

#ifndef COUNT_COLUMNS
#define COUNT_COLUMNS 0
#endif

#ifndef COUNT_COLUMN_OPERATIONS
#define COUNT_COLUMN_OPERATIONS 0
#endif

#ifndef COUNT_COLUMN_TIME
#define COUNT_COLUMN_TIME 0
#endif

#ifndef COUNT_ADDITIONS
#define COUNT_ADDITIONS 0
#endif

#ifndef WITH_KILLING
#define WITH_KILLING 1
#endif

#ifndef WITH_FLIPPING
#define WITH_FLIPPING 0
#endif

#ifndef WRITE_IN_LINBOX
#define WRITE_IN_LINBOX 0
#endif

#ifndef PRINT_MATRICES
#define PRINT_MATRICES 1
#endif 

#define BY_ROWS 0

#ifndef USE_VECTOR_FOR_COLUMNS
#define USE_VECTOR_FOR_COLUMNS 1
#endif

#if COUNT_ADDITIONS
unsigned long add_basic=0;
#endif

typedef         unsigned long	                    Vertex;
typedef         unsigned long                       Index;

#if USE_DIONYSUS_SIMPLEX
typedef         Simplex<Vertex, float>              Smplx;
#else 
typedef std::pair<std::list<Vertex>,float>            Smplx;
#endif


template <typename Simplex_type> struct Simplex_traits {
};

#if USE_DIONYSUS_SIMPLEX
template <typename VertexType,typename DataType> struct Simplex_traits<Simplex<VertexType,DataType> > {

    typedef VertexType Vertex_type;
    typedef DataType Data_type;

    typedef Simplex<Vertex_type,Data_type> Smplx;

    struct Dimension {
        int operator() (const Smplx& s) {
            return s.dimension();
        }
    };

    struct Data {
        Data_type operator() (const Smplx& s) {
            return s.data();
        }
    };

    struct Set_data {
        void operator() (Smplx& s,const Data_type& d) {
            s.data()=d;
        }
    };

    struct Add_vertex {
        void operator() (Smplx& s, const Vertex_type& v) {
            s.add(v);
        }
    };

    struct Is_coned {

        Index death;
        bool has_death;

        Is_coned() : has_death(false) {}
        Is_coned(Index death) : death(death), has_death(true) {}

        bool operator() (const Smplx& s) {
            //std::cerr << "is coned: " << s << " - " << death << std::endl;
            return has_death && s.contains(death);
        }
    };  

    typedef typename Smplx::VertexComparison Comparison;

    typedef typename Smplx::BoundaryIterator Boundary_iterator;

    struct Boundary_begin {
        Boundary_iterator operator() (const Smplx& s) {
            return s.boundary_begin();
        }
    };

    struct Boundary_end {
        Boundary_iterator operator() (const Smplx& s) {
            return s.boundary_end();
        }
    };

};

#endif

template <typename VertexType,typename DataType> struct Simplex_traits<std::pair<std::list<VertexType>,DataType> > {

    typedef VertexType Vertex_type;
    typedef DataType Data_type;

    typedef std::pair<std::list<VertexType>,DataType> Smplx;

    struct Dimension {
        int operator() (const Smplx& s) {
            return s.first.size()-1;
        }
    };

    struct Data {
        Data_type operator() (const Smplx& s) {
            return s.second;
        }
    };

    struct Set_data {
        void operator() (Smplx& s,const Data_type& d) {
            s=make_pair(s.first,d);
        }
    };

    struct Add_vertex {
        void operator() (Smplx& s, const Vertex_type& v) {
            s.first.insert(s.first.end(),v);
        }
    };

    struct Is_coned {

        Index death;
        bool has_death;

        Is_coned() : has_death(false) {}
        Is_coned(Index death) : death(death), has_death(true) {}

        bool operator() (const Smplx& s) {
            //std::cerr << "is coned: " << s << " - " << death << std::endl;
            return has_death && s.first.back()==death;
        }
    };  

    struct Comparison {
        bool operator() (const Smplx& s1,const Smplx& s2) const {
            if(s1.first.size()!=s2.first.size()) {
                return s1.first.size()<s2.first.size();
            }
            typename Smplx::first_type::const_iterator it1=s1.first.begin(),
                it2=s2.first.begin();
            while(it1!=s1.first.end()) {
                if(*it1!=*it2) {
                    return *it1<*it2;
                }
                it1++;
                it2++;
            }
            return false;
        }
    };



    struct Boundary_iterator {

        Index index;

        const Smplx& s;

        Boundary_iterator(const Smplx& s,bool at_end) : s(s) {
            if(at_end) {
                index = s.first.size();
                if(Dimension()(s)==0) {
                    index--;
                }
            } else {
                index =0;
            }
        }

        void operator++ () {
            index++;
        }

        Smplx operator* () {
            Smplx boundary;
            Add_vertex add;
            typename std::list<Vertex_type>::const_iterator it=s.first.begin();
            for(Index i=0;i<s.first.size();i++,it++) {
                if(i!=index) {
                    add(boundary,*it);
                }
            }
            return boundary;
        }

        bool operator== (const Boundary_iterator& it) {
            return index==it.index;
        }

        bool operator!= (const Boundary_iterator& it) {
            return !this->operator== (it);
        }

    };

    struct Boundary_begin {
        Boundary_iterator operator() (Smplx& s) {
            return Boundary_iterator(s,false);
        }
    };

    struct Boundary_end {
        Boundary_iterator operator() (Smplx& s) {
            return Boundary_iterator(s,true);
        }
    };


    /*

    struct Boundary_iterator {

    Index index;

    Smplx& s;
    typename Smplx::first_type::iterator curr_pos;
    Vertex_type last_deleted_val;
    bool at_end;

    Boundary_iterator(Smplx& s,bool at_end) : s(s), at_end(at_end) {
    if(!at_end) {
    last_deleted_val=s.first.front();
    curr_pos=s.first.erase(s.first.begin());
    this->at_end=s.first.empty();
    if(this->at_end) {
    std::cout << "Now at end!" << std::endl;
    }
    }
    }

    void operator++ () {
    s.first.insert(curr_pos,last_deleted_val);
    if(curr_pos!=s.first.end()) {
    last_deleted_val=*curr_pos;
    } else {
    at_end=true;
    }
    curr_pos=s.first.erase(curr_pos);
    }

    Smplx& operator* () {
    return s;
    }

    bool operator== (const Boundary_iterator& it) {
    if(at_end || it.at_end) {
    if(! (it.at_end && at_end)) {
    return false;
    }
    return true;
    }
    return last_deleted_val==it.last_deleted_val;
    }

    bool operator!= (const Boundary_iterator& it) {
    return !this->operator== (it);
    }

    };

    struct Boundary_begin {
    Boundary_iterator operator() (Smplx& s) {
    return Boundary_iterator(s,false);
    }
    };

    struct Boundary_end {
    Boundary_iterator operator() (Smplx& s) {
    return Boundary_iterator(s,true);
    }
    };

    */

};

typedef         std::map<Smplx, Index, 
    Simplex_traits<Smplx>::Comparison>   Complex;
typedef         std::vector<Smplx>                  SmplxVector;


//typedef          std::set<Index>                     Column;
#if USE_VECTOR_FOR_COLUMNS
typedef          std::vector<Index>                     Column;
#else
typedef          std::list<Index>                     Column;
#endif

typedef std::vector<Column> Matrix;



template <typename ColumnType> struct Column_traits {

    typedef ColumnType Column;

    typedef void Index;

    struct Is_zero {
        bool operator() (Column& c) {
            return false;
        }
    };

    struct Set_zero {
        void operator() (Column& c) {
        }
    };

    struct Lowest{
        Index operator()(const Column& c) {
            //return 0;
        }
    };

    struct Add_to {
        const Column& operator()(Column& c,const Column& d) {
            return c;
        }
    };

};

template<typename Index_> struct Column_traits<std::set<Index_> > {

    typedef Index_ Index;

    typedef std::set<Index> Column;

    typedef typename Column::iterator Column_iterator;

    struct Construct {
        template<typename InputIterator> void operator() 
            (InputIterator begin, InputIterator end,Column& c) const {
                c.insert(begin,end);
        }
    };


    struct Print {
        void operator() (const Column& c) const {
            for(Column_iterator it=c.begin();it!=c.end();it++) {
                std::cout << *it << " " << std::flush;
            }
        }
    };

    struct Is_zero {
        bool operator() (const Column& c) const {
            return c.empty();
        }
    };

    struct Set_zero {
        void operator() (Column& c) {
            c.clear();
        }
    };

    struct Lowest{
        Index operator()(const Column& c) const {
            assert(!Is_zero()(c));
            return *(c.rbegin());
        }
    };

    struct Length {
        Index operator()(const Column& c) const {
            return c.size();
        }
    };


    struct Add_to {
        const Column& operator()(Column& c,const Column& d) {
            assert(!Is_zero()(c));
            assert(!Is_zero()(d));
            for(Column_iterator it=d.begin(); it!=d.end();it++) {
                Column_iterator lookup=c.lower_bound(*it);
                if(lookup!=c.end() && *lookup==*it) { // element found
                    c.erase(lookup);
                } else { // we have to insert
                    if(lookup!=c.begin()) {
                        lookup--;
                        c.insert(lookup,*it);
                    } else {
                        c.insert(*it);
                    }
                }
            }
            return c;
        }
    };

};

template<typename Index_> struct Column_traits<std::list<Index_> > {

    typedef Index_ Index;

    typedef std::list<Index> Column;

    typedef typename Column::iterator Column_iterator;
    typedef typename Column::const_iterator Column_const_iterator;

    typedef Column_traits<Column> Self;

    struct Construct {
        template<typename InputIterator> void operator() 
            (InputIterator begin, InputIterator end,Column& c) const {
                std::sort(begin,end);
                c.insert(c.begin(),begin,end);
        }
    };


    struct Print {
        void operator() (const Column& c) const {
            for(Column_const_iterator it=c.begin();it!=c.end();it++) {
                std::cout << *it << " " << std::flush;
            }
        }
    };

    struct Is_zero {
        bool operator() (const Column& c) const {
            return c.empty();
        }
    };

    struct Set_zero {
        void operator() (Column& c) {
            c.clear();
        }
    };

    struct Lowest{
        Index operator()(const Column& c) const {
            assert(!Is_zero()(c));
            assert(*(std::max_element(c.begin(),c.end()))==c.back());
            return c.back();
        }
    };

    struct Length {
        Index operator()(const Column& c) const {
            return c.size();
        }
    };


    struct Add_to {
        const Column& operator()(Column& c,const Column& d) {
            //Self::Print print;
            //std::cout << "len(c)=" << Length()(c)<< std::endl;
            //std::cout << "len(d)=" << Length()(d) << std::endl;
            assert(!Is_zero()(c));
            assert(!Is_zero()(d));
            Column_iterator c_it=c.begin();
            Column_const_iterator d_it=d.begin();
            while(d_it!=d.end()) {
                if(c_it==c.end()) {
                    c.insert(c_it,d_it,d.end());
#if COUNT_ADDITIONS
                    add_basic+=std::distance(d_it,d.end());
#endif
                    break;
                }
                while(*d_it>*c_it) {
                    c_it++;
#if COUNT_ADDITIONS
                    add_basic++;
#endif
                }
                if(*d_it==*c_it) {
                    c_it=c.erase(c_it);
                } else {
                    c.insert(c_it,*d_it);
                }
                d_it++;
#if COUNT_ADDITIONS
                add_basic++;
#endif

            }
            //std::cout << "r=";print(c);std::cout << std::endl;
            return c;
        }
    };

};

template<typename Index_> struct Column_traits<std::vector<Index_> > {

    typedef Index_ Index;

    typedef std::vector<Index> Column;

    typedef typename Column::iterator Column_iterator;
    typedef typename Column::const_iterator Column_const_iterator;

    typedef Column_traits<Column> Self;

    struct Construct {
        template<typename InputIterator> void operator() 
            (InputIterator begin, InputIterator end,Column& c) const {
                std::sort(begin,end);
                c.insert(c.begin(),begin,end);
        }
    };


    struct Print {
        void operator() (const Column& c) const {
            for(Column_const_iterator it=c.begin();it!=c.end();it++) {
                std::cout << *it << " " << std::flush;
            }
        }
    };

    struct Is_zero {
        bool operator() (const Column& c) const {
            return c.empty();
        }
    };

    struct Set_zero {
        void operator() (Column& c) {
            c.clear();
        }
    };

    struct Lowest{
        Index operator()(const Column& c) const {
            assert(!Is_zero()(c));
            assert(*(std::max_element(c.begin(),c.end()))==c.back());
            return c.back();
        }
    };

    struct Length {
        Index operator()(const Column& c) const {
            return c.size();
        }
    };


    struct Add_to {
        const Column& operator()(Column& c,const Column& d) {
            assert(!Is_zero()(c));
            assert(!Is_zero()(d));
            Column_iterator c_it=c.begin();
            Column_const_iterator d_it=d.begin();
            Column new_c;
            new_c.reserve(d.size());
            while(c_it!=c.end() || d_it!=d.end()) {
                if(c_it==c.end()) {
                    new_c.insert(new_c.end(),d_it,d.end());
#if COUNT_ADDITIONS
                    add_basic+=std::distance(d_it,d.end());
#endif
                    break;
                }
                if(d_it==d.end()) {
                    new_c.insert(new_c.end(),c_it,c.end());
#if COUNT_ADDITIONS
                    add_basic+=std::distance(c_it,c.end());
#endif
                    break;
                }
                if(*d_it==*c_it) {
                    c_it++;
                    d_it++;
                } else if(*d_it<*c_it) {
                    new_c.push_back(*d_it);
                    d_it++;
                } else {
                    new_c.push_back(*c_it);
                    c_it++;
                }
#if COUNT_ADDITIONS
                add_basic++;
#endif
            }
            //std::cout << "r=";print(c);std::cout << std::endl;
            c=new_c;
            return c;
        }
    };

};

#if PRINT_MATRICES

void print_matrix(const Matrix& m) {
    Column_traits<Column>::Print print;
    Simplex_traits<Smplx>::Dimension dim;
    std::cout << "BEGIN MATRIX" << std::endl;
    std::cout << "Matrix of size " << m.size() << std::endl;
    for(std::size_t i=0;i<m.size();i++) {
        std::cout << "Column " << i << ": ";
        print(m[i]);        
        std::cout << std::endl;
    }
    std::cout << "END MATRIX" << std::endl;
}

#endif

/*#if PRINT_MATRICES

void print_matrix(const Matrix& m) {
std::ofstream ofs("matrix.txt");
Column_traits<Column>::Print print;
Simplex_traits<Smplx>::Dimension dim;
ofs << "BEGIN MATRIX" << std::endl;
ofs << "Matrix of size " << m.size() << std::endl;
for(std::size_t i=0;i<m.size();i++) {
ofs << "Column " << i << ": ";
print(m[i]);        
ofs << "\n";
}
ofs << "END MATRIX" << std::endl;
}

#endif*/

template<typename Map>
struct Iterator_to_map {

    typedef typename Map::mapped_type result_type;

    Iterator_to_map(Map* map) : _map(map) {}

    typename Map::mapped_type operator() (const typename Map::key_type& key) 
        const {
            return (*_map)[key];
    }

private:
    Map* _map;

};

void randomized_reduce(Matrix& R) {

    typedef Column_traits<Column> Column_traits;
    Column_traits::Is_zero is_zero;
    Column_traits::Set_zero set_zero;
    Column_traits::Add_to add_to;
    Column_traits::Lowest lowest;

    Index n = R.size();

    for(Index j=0;j<n;j++) {
        if(!is_zero(R[j])) {
            Index i = lowest(R[j]);
            if(!is_zero(R[i])) {
                set_zero(R[i]);
            }
        }
    }

    std::set<Index> S;
    std::vector<std::set<Index> > L(n);

    typedef std::set<Index>::iterator Set_iterator;
    Set_iterator fast_insert;

    for(Index j=0;j<n;j++) {
        if(!is_zero(R[j])) {
            Index i = lowest(R[j]);
            if(L[i].empty()) {
                L[i].insert(j);
            } else {
                fast_insert = L[i].end();
                fast_insert--;
                L[i].insert(fast_insert,j);
                if(S.empty()) {
                    S.insert(j);
                } else {
                    fast_insert=S.end();
                    fast_insert--;
                    S.insert(fast_insert,j);
                }
            }
        }
    }

#if COUNT_COLUMN_OPERATIONS
    Column_traits::Length length;
    std::vector<Index> col_ops(n), col_ops_weighted(n);
    for(Index i=0;i<n;i++) {
        col_ops[i]=col_ops_weighted[i]=0;
    }
#endif

#if COUNT_COLUMN_TIME
    std::map<Index,clock_t> col_time;
    for(Index i=0;i<n;i++) {
        col_time[i]=clock_t(0);
    }
#endif


    srand(time(NULL));

    Set_iterator j_in_s=S.begin(),next_j_in_s;

    while(! S.empty()) {

        /*
        for(Index i=0;i<n;i++) {
        if(L[i].empty()) {
        continue;
        }
        Set_iterator it=L[i].begin();
        assert(S.find(*it)==S.end());
        while(++it!=L[i].end()) {
        assert(S.find(*it)!=S.end());
        }
        }
        */    

        /*
        Index el_in_s = S.size();
        Index steps = rand() % el_in_s;
        */
        next_j_in_s=j_in_s;
        next_j_in_s++;
        Index j = *j_in_s;

        assert(!is_zero(R[j]));
        Index lowest_j = lowest(R[j]);
        Index k = *(L[lowest_j].begin());
        assert(j!=k);
        assert(lowest(R[k])==lowest(R[j]));
#if COUNT_COLUMN_TIME
        clock_t stamp=clock();
#endif
        add_to(R[j],R[k]);
#if COUNT_COLUMN_TIME
        col_time[j]+=clock()-stamp;
#endif
#if COUNT_COLUMN_OPERATIONS
        col_ops[j]++;
        col_ops_weighted[j]+=length(R[k]);
#endif
        L[lowest_j].erase(j);
        if(is_zero(R[j])) {
            assert(*j_in_s==j);
            S.erase(j_in_s);
        } else {
            Index i = lowest(R[j]);
            if(L[i].empty()) {
                if(!is_zero(R[i])) {
                    Index ell=lowest(R[i]);
                    assert(L[ell].find(i)!=L[ell].end());
                    L[ell].erase(i);
                    if(!L[ell].empty()) {
                        Index tmp=*(L[ell].begin());
                        if(i < tmp) {
                            assert(S.find(tmp)!=S.end());
                            S.erase(tmp);
                        } else {
                            assert(S.find(i)!=S.end());
                            S.erase(i);
                        }
                    }
                    set_zero(R[i]);
                }
                assert(*j_in_s==j);
                S.erase(j_in_s);
            }
            if(!L[i].empty() && j < *(L[i].begin())) {
                assert(*j_in_s==j);
                S.erase(j_in_s);
                S.insert(*(L[i].begin()));
            }
            L[i].insert(j);
        }
        j_in_s=next_j_in_s;
        if(j_in_s==S.end()) {
            j_in_s=S.begin();
        }
    }


#if COUNT_COLUMNS
    Index col_neg=0, col_pos_paired=0,col_essential=0;
    for(Index i=0;i<n;i++) {
        if(! is_zero(R[i])) {
            col_neg++;
        } else {
            if(L[i].empty()) {
                col_essential++;
            } else {
                col_pos_paired++;
            }
        }
    }
    std::cout << "Column: " << std::endl;
    std::cout << "negative   : " << col_neg << std::endl;
    std::cout << "pos paired : " << col_pos_paired << std::endl;
    std::cout << "essential  : " << col_essential << std::endl;
#endif

#if COUNT_COLUMN_OPERATIONS
    Index col_ops_neg=0, col_ops_pos_paired=0,col_ops_essential=0;
    Index col_ops_neg_weighted=0, col_ops_pos_paired_weighted=0,
        col_ops_essential_weighted=0;
    for(Index i=0;i<n;i++) {
        if(! is_zero(R[i])) {
            col_ops_neg+=col_ops[i];
            col_ops_neg_weighted+=col_ops_weighted[i];
        } else {
            if(L[i].empty()) {
                col_ops_essential+=col_ops[i];
                col_ops_essential_weighted+=col_ops_weighted[i];
            } else {
                col_ops_pos_paired+=col_ops[i];
                col_ops_pos_paired_weighted+=col_ops_weighted[i];
            }
        }
    }
    std::cout << "Column operations: " << std::endl;
    std::cout << "On negative columns: " << col_ops_neg << " (" << col_ops_neg_weighted << ")" << std::endl;
    std::cout << "On pos paired cols : " << col_ops_pos_paired << " (" << col_ops_pos_paired_weighted << ")"<< std::endl;
    std::cout << "On essential cols  : " << col_ops_essential << " (" << col_ops_essential_weighted << ")" << std::endl;
#endif
#if COUNT_COLUMN_TIME
    clock_t col_time_neg=0, col_time_pos_paired=0,col_time_essential=0;
    for(Index i=0;i<n;i++) {
        if(! is_zero(R[i])) {
            col_time_neg+=col_time[i];
        } else {
            if(L[i].empty()) {
                col_time_essential+=col_time[i];
            } else {
                col_time_pos_paired+=col_time[i];
            }
        }
    }
    std::cout << "Column time: " << std::endl;
    std::cout << "On negative columns: " << double(col_time_neg)/CLOCKS_PER_SEC << std::endl;
    std::cout << "On pos paired cols : " << double(col_time_pos_paired)/CLOCKS_PER_SEC << std::endl;
    std::cout << "On essential cols  : " << double(col_time_essential)/CLOCKS_PER_SEC << std::endl;
#endif

}

void deterministic_reduce(Matrix& R) {

    typedef Column_traits<Column> Column_traits;
    Column_traits::Is_zero is_zero;
    Column_traits::Add_to add_to;
    Column_traits::Lowest lowest;

    Index n = R.size();

    std::vector<Index> L(n);

    typedef std::set<Index>::iterator Set_iterator;
    Set_iterator fast_insert;

#if COUNT_COLUMN_OPERATIONS
    Column_traits::Length length;
    std::vector<Index> col_ops(n), col_ops_weighted(n);
    for(Index i=0;i<n;i++) {
        col_ops[i]=col_ops_weighted[i]=0;
    }
#endif

#if COUNT_COLUMN_TIME
    std::map<Index,clock_t> col_time;
    for(Index i=0;i<n;i++) {
        col_time[i]=clock_t(0);
    }
#endif

    for(Index j=0;j<n;j++) {
        L[j]=0;
    }
    srand(time(NULL));

    for(Index j=0;j<n;j++) {
        while(!is_zero(R[j]) && L[lowest(R[j])]!=0) {
            Index k = L[lowest(R[j])];
#if COUNT_COLUMN_TIME
            clock_t stamp=clock();
#endif
            add_to(R[j],R[k]);
#if COUNT_COLUMN_TIME
            col_time[j]+=clock()-stamp;
#endif
#if COUNT_COLUMN_OPERATIONS
            col_ops[j]++;
            col_ops_weighted[j]+=length(R[k]);
#endif
        }

        if(!is_zero(R[j])) {
            L[lowest(R[j])]=j;
        }
    }

#if COUNT_COLUMNS
    Index col_neg=0, col_pos_paired=0,col_essential=0;
    for(Index i=0;i<n;i++) {
        if(! is_zero(R[i])) {
            col_neg++;
        } else {
            if(L[i]==0) {
                col_essential++;
            } else {
                col_pos_paired++;
            }
        }
    }

    std::cout << "Column: " << std::endl;
    std::cout << "negative   : " << col_neg << std::endl;
    std::cout << "pos paired : " << col_pos_paired << std::endl;
    std::cout << "essential  : " << col_essential << std::endl;
#endif

#if COUNT_COLUMN_OPERATIONS
    Index col_ops_neg=0, col_ops_pos_paired=0,col_ops_essential=0;
    Index col_ops_neg_weighted=0, col_ops_pos_paired_weighted=0,
        col_ops_essential_weighted=0;
    for(Index i=0;i<n;i++) {
        if(! is_zero(R[i])) {
            col_ops_neg+=col_ops[i];
            col_ops_neg_weighted+=col_ops_weighted[i];
        } else {
            if(L[i]==0) {
                col_ops_essential+=col_ops[i];
                col_ops_essential_weighted+=col_ops_weighted[i];
            } else {
                col_ops_pos_paired+=col_ops[i];
                col_ops_pos_paired_weighted+=col_ops_weighted[i];
            }
        }
    }
    std::cout << "Column operations: " << std::endl;
    std::cout << "On negative columns: " << col_ops_neg << " (" << col_ops_neg_weighted << ")" << std::endl;
    std::cout << "On pos paired cols : " << col_ops_pos_paired << " (" << col_ops_pos_paired_weighted << ")"<< std::endl;
    std::cout << "On essential cols  : " << col_ops_essential << " (" << col_ops_essential_weighted << ")" << std::endl;
#endif
#if COUNT_COLUMN_TIME
    clock_t col_time_neg=0, col_time_pos_paired=0,col_time_essential=0;
    for(Index i=0;i<n;i++) {
        if(! is_zero(R[i])) {
            col_time_neg+=col_time[i];
        } else {
            if(L[i]==0) {
                col_time_essential+=col_time[i];
            } else {
                col_time_pos_paired+=col_time[i];
            }
        }
    }
    std::cout << "Column time: " << std::endl;
    std::cout << "On negative columns: " << col_time_neg << std::endl;
    std::cout << "On pos paired cols : " << col_time_pos_paired << std::endl;
    std::cout << "On essential cols  : " << col_time_essential << std::endl;
#endif

}

void deterministic_reduce_by_rows(Matrix& R) {

    typedef Column_traits<Column> Column_traits;
    Column_traits::Is_zero is_zero;
    Column_traits::Set_zero set_zero;
    Column_traits::Add_to add_to;
    Column_traits::Lowest lowest;

    Index n = R.size();

    typedef std::priority_queue<Index,std::vector<Index>, std::greater<Index> > PQ;

    std::vector<PQ> indices(n);

    std::vector<bool> killed;

    typedef std::set<Index>::iterator Set_iterator;
    Set_iterator fast_insert;

#if COUNT_COLUMN_OPERATIONS
    Column_traits::Length length;
    std::vector<Index> col_ops(n), col_ops_weighted(n);
    for(Index i=0;i<n;i++) {
        col_ops[i]=col_ops_weighted[i]=0;
    }
#endif

#if COUNT_COLUMN_TIME
    std::map<Index,clock_t> col_time;
    for(Index i=0;i<n;i++) {
        col_time[i]=clock_t(0);
    }
#endif

    for(Index j=0;j<n;j++) {
        if(! is_zero(R[j])) {
            indices[lowest(R[j])].push(j);
        }
        killed.push_back(false);
    }

    for(long i = n-1; i >=0; i--) {
        PQ::size_type m = indices[i].size();
        if(m==0) {
            continue;
        }
        Index p = indices[i].top();
#if WITH_KILLING
        killed[lowest(R[p])]=true;
#endif
        indices[i].pop();
        while(!indices[i].empty()) {
            Index next = indices[i].top();
            indices[i].pop();
#if COUNT_COLUMN_TIME
            clock_t stamp=clock();
#endif
            if(killed[next]) {
                set_zero(R[next]);
            } else {
                add_to(R[next],R[p]);
                if(! is_zero(R[next])) {
                    indices[lowest(R[next])].push(next);
                }      
            }

#if COUNT_COLUMN_TIME
            col_time[next]+=clock()-stamp;
#endif
#if COUNT_COLUMN_OPERATIONS
            col_ops[j]++;
            col_ops_weighted[j]+=length(R[k]);
#endif
        }
    }
#if COUNT_COLUMNS
    Index col_neg=0, col_pos_paired=0,col_essential=0;
    for(Index i=0;i<n;i++) {
        if(! is_zero(R[i])) {
            col_neg++;
        } else {
            if(L[i]==0) {
                col_essential++;
            } else {
                col_pos_paired++;
            }
        }
    }

    std::cout << "Column: " << std::endl;
    std::cout << "negative   : " << col_neg << std::endl;
    std::cout << "pos paired : " << col_pos_paired << std::endl;
    std::cout << "essential  : " << col_essential << std::endl;
#endif

#if COUNT_COLUMN_OPERATIONS
    Index col_ops_neg=0, col_ops_pos_paired=0,col_ops_essential=0;
    Index col_ops_neg_weighted=0, col_ops_pos_paired_weighted=0,
        col_ops_essential_weighted=0;
    for(Index i=0;i<n;i++) {
        if(! is_zero(R[i])) {
            col_ops_neg+=col_ops[i];
            col_ops_neg_weighted+=col_ops_weighted[i];
        } else {
            if(L[i]==0) {
                col_ops_essential+=col_ops[i];
                col_ops_essential_weighted+=col_ops_weighted[i];
            } else {
                col_ops_pos_paired+=col_ops[i];
                col_ops_pos_paired_weighted+=col_ops_weighted[i];
            }
        }
    }
    std::cout << "Column operations: " << std::endl;
    std::cout << "On negative columns: " << col_ops_neg << " (" << col_ops_neg_weighted << ")" << std::endl;
    std::cout << "On pos paired cols : " << col_ops_pos_paired << " (" << col_ops_pos_paired_weighted << ")"<< std::endl;
    std::cout << "On essential cols  : " << col_ops_essential << " (" << col_ops_essential_weighted << ")" << std::endl;
#endif
#if COUNT_COLUMN_TIME
    clock_t col_time_neg=0, col_time_pos_paired=0,col_time_essential=0;
    for(Index i=0;i<n;i++) {
        if(! is_zero(R[i])) {
            col_time_neg+=col_time[i];
        } else {
            col_time_essential+=col_time[i];
        }
    }
    std::cout << "Column time: " << std::endl;
    std::cout << "On negative columns: " << col_time_neg << std::endl;
    std::cout << "On pos paired cols : " << col_time_pos_paired << std::endl;
    std::cout << "On essential cols  : " << col_time_essential << std::endl;
#endif

}



void deterministic_clever_reduce(Matrix& R) {

    typedef Matrix::value_type Column;

    typedef Column_traits<Column> Column_traits;
    Column_traits::Is_zero is_zero;
    Column_traits::Set_zero set_zero;
    Column_traits::Add_to add_to;
    Column_traits::Lowest lowest;
    Column_traits::Length length;

    Index n = R.size();

    std::vector<Index> L(n);

#if COUNT_COLUMN_OPERATIONS
    std::vector<Index> col_ops(n), col_ops_weighted(n);
    for(Index i=0;i<n;i++) {
        col_ops[i]=col_ops_weighted[i]=0;
    }
#endif

#if COUNT_COLUMN_TIME
    std::map<Index,clock_t> col_time;
    for(Index i=0;i<n;i++) {
        col_time[i]=clock_t(0);
    }
#endif

    Index max_d=0;

    std::vector<std::vector<Index> > paths;

    for(Index j=0;j<n;j++) {
        L[j]=0;
        int dim_of_col = length(R[j]) - 1;
        if(dim_of_col==-1) {
            continue;
        }
        if(dim_of_col>max_d) {
            max_d=dim_of_col;
            paths.resize(max_d+1);
        }
        paths[dim_of_col].push_back(j);    
    }

    for(Index d=max_d;d>=1;d--) {
        for(std::vector<Index>::iterator it=paths[d].begin();it!=paths[d].end();it++) {
            Index j = *it;
            while(!is_zero(R[j]) && L[lowest(R[j])]!=0) {
                Column& R_k = R[L[lowest(R[j])]];
#if COUNT_COLUMN_TIME
                clock_t stamp=clock();
#endif
                add_to(R[j],R_k);
#if COUNT_COLUMN_TIME
                col_time[j]+=clock()-stamp;
#endif
#if COUNT_COLUMN_OPERATIONS
                col_ops[j]++;
                col_ops_weighted[j]+=length(R_k);
#endif
            }

            if(!is_zero(R[j])) {
                Index k = lowest(R[j]);
                L[k]=j;
                set_zero(R[k]);
            }
        }
    }

#if COUNT_COLUMNS
    Index col_neg=0, col_pos_paired=0,col_essential=0;
    for(Index i=0;i<n;i++) {
        if(! is_zero(R[i])) {
            col_neg++;
        } else {
            if(L[i]==0) {
                col_essential++;
            } else {
                col_pos_paired++;
            }
        }
    }
    std::cerr << "Column: " << std::endl;
    std::cerr << "negative   : " << col_neg << std::endl;
    std::cerr << "pos paired : " << col_pos_paired << std::endl;
    std::cerr << "essential  : " << col_essential << std::endl;
#endif

#if COUNT_COLUMN_OPERATIONS
    Index col_ops_neg=0, col_ops_pos_paired=0,col_ops_essential=0;
    Index col_ops_neg_weighted=0, col_ops_pos_paired_weighted=0,
        col_ops_essential_weighted=0;
    for(Index i=0;i<n;i++) {
        if(! is_zero(R[i])) {
            col_ops_neg+=col_ops[i];
            col_ops_neg_weighted+=col_ops_weighted[i];
        } else {
            if(L[i]==0) {
                col_ops_essential+=col_ops[i];
                col_ops_essential_weighted+=col_ops_weighted[i];
            } else {
                col_ops_pos_paired+=col_ops[i];
                col_ops_pos_paired_weighted+=col_ops_weighted[i];
            }
        }
    }
    std::cerr << "Column operations: " << std::endl;
    std::cerr << "On negative columns: " << col_ops_neg << " (" << col_ops_neg_weighted << ")" << std::endl;
    std::cerr << "On pos paired cols : " << col_ops_pos_paired << " (" << col_ops_pos_paired_weighted << ")"<< std::endl;
    std::cerr << "On essential cols  : " << col_ops_essential << " (" << col_ops_essential_weighted << ")" << std::endl;
#endif
#if COUNT_COLUMN_TIME
    clock_t col_time_neg=0, col_time_pos_paired=0,col_time_essential=0;
    for(Index i=0;i<n;i++) {
        if(! is_zero(R[i])) {
            col_time_neg+=col_time[i];
        } else {
            if(L[i]==0) {
                col_time_essential+=col_time[i];
            } else {
                col_time_pos_paired+=col_time[i];
            }
        }
    }
    std::cerr << "Column time: " << std::endl;
    std::cerr << "On negative columns: " << col_time_neg << std::endl;
    std::cerr << "On pos paired cols : " << col_time_pos_paired << std::endl;
    std::cerr << "On essential cols  : " << col_time_essential << std::endl;
#endif

}

Matrix anti_transpose(const Matrix& R) {

    int n = R.size();
    Matrix S(n);

    for(int i=0;i<n;i++) {
        for(int j=0;j<R[i].size();j++) {
            S[n-R[i][j]-1].push_back(n-i-1);
        }
    }
    for(int i=0;i<n;i++) {
        std::sort(S[i].begin(),S[i].end());
    }

    return S;

}

#if WRITE_IN_LINBOX
void write_in_linbox_format(const Matrix& R, std::ostream& ostr) {

    int n = R.size();
    Matrix S(n);

    for(int i=0;i<n;i++) {
        for(int j=0;j<R[i].size();j++) {
            S[R[i][j]].push_back(i);
        }
    }
    for(int i=0;i<n;i++) {
        std::sort(S[i].begin(),S[i].end());
    }

    ostr << n << " " << n << " S" << std::endl;
    for(int i=0;i<n;i++) {
        ostr << S[i].size() << " ";
        for(int j=0;j<S[i].size();j++) {
            ostr << S[i][j] << " 1 ";
        }
        std::cout << std::endl;
    }
    return;

}
#endif

//void handle(char *inFileName)
void handle() {

    // ... open the input file
    //std::cerr << inFileName << std::endl;
    //std::ifstream inFile(inFileName);
    //std::string str(inFileName);
    //std::ofstream ofs(str.replace(str.find("txt"),3,"dgm").c_str());

    /*if (! inFile.is_open())
    {
        std::cerr << "Could not open file, " << inFileName << std::endl;
        exit(-1);
	}*/

    Index death;

    float curr_val;

    Complex         c;

    SmplxVector v;

    Matrix R;

    long no_regular=0, no_extended=0, no_relative=0;
    long no_dim[4] = { 0, 0, 0, 0 };
    double total_persistence[3]={0., 0., 0.};

    Iterator_to_map<Complex> it_to_map(&c);

    typedef Column_traits<Column> Column_traits;
    Column_traits::Construct construct;
    Column_traits::Is_zero is_zero;
    Column_traits::Lowest lowest;

    Simplex_traits<Smplx>::Dimension dim;
    Simplex_traits<Smplx>::Data data;
    Simplex_traits<Smplx>::Set_data set_data;
    Simplex_traits<Smplx>::Add_vertex add_vertex;
    Simplex_traits<Smplx>::Is_coned is_coned;

    typedef Simplex_traits<Smplx>::Boundary_iterator Boundary_iterator;
    Simplex_traits<Smplx>::Boundary_begin boundary_begin;
    Simplex_traits<Smplx>::Boundary_end boundary_end;

    boost::timer overall_timer;

    //while(inFile.good())
    while(std::cin) {
        char curr[4096];
	std::cin.getline(curr,4096);
        std::string str(curr);
        //std::cerr << str << std::endl;
        if(str=="") {
            continue;
        }
        std::stringstream sstr(str);	
        char indicator;
        sstr >> indicator;
        if(indicator=='t') {
            std::string th;
            sstr >> th;
            overall_timer.restart();      
        }
        if(indicator=='d') {
            sstr >> death;
            is_coned = Simplex_traits<Smplx>::Is_coned(death);
        }
        if(indicator=='v') {
            sstr >> curr_val;
        }
        if(indicator=='!') {
            std::cerr << "Time for reading simplices: " << overall_timer.elapsed() << std::endl;
            overall_timer.restart();
            Index n = v.size();
            R.reserve(n);
            Column col; //dummy
            for(Index i=0;i<n;i++) {
                //std::cerr << "-------" << smpl << std::endl;
                Smplx& smpl=v[i];

                std::vector<Index>      boundary;
                for (Boundary_iterator bcur  = boundary_begin(smpl); 
                    bcur != boundary_end(smpl);       ++bcur) {
                        const Smplx& bcurr=*bcur;
                        assert(c.find(bcurr)!=c.end());
                        boundary.push_back(c[bcurr]);
                }
                R.push_back(col);
                construct(boundary.begin(),boundary.end(),R.back());
                c[smpl] = i;
            }

            std::cerr << "Time for constructing boundary matrix of size " << R.size() << ": " << overall_timer.elapsed() << std::endl;
//#if PRINT_MATRICES
//            print_matrix(R);
//#endif
//            std::cerr << "Reduce.." << std::flush;
//            overall_timer.restart();
//
//#if WRITE_IN_LINBOX
//            write_in_linbox_format(R,std::cout);
//            std::exit(1);
//#endif
//
//#if WITH_FLIPPING
//            std::cerr << "Taking anti-transpose" << std::endl;
//            R = anti_transpose(R);
//            std::cerr << "Time for antitranspose" << ": " << overall_timer.elapsed() << std::endl;
//            std::cerr << "Reduce.." << std::flush;
//            overall_timer.restart();
//#endif
//
//#if BY_ROWS
//            deterministic_reduce_by_rows(R);
//#else
//#if WITH_KILLING
//            deterministic_clever_reduce(R);
//#else
//            deterministic_reduce(R);
//#endif
//#endif
            

            std::vector< phat::dimension > dims( R.size() );
            for( std::size_t idx = 0; idx < R.size(); idx++ )
                dims[ idx ] = std::max( R[ idx ].size() - 1, (std::size_t)0 );
            phat::boundary_matrix< phat::bit_tree_pivot_column > boundary_matrix;
            boundary_matrix.load_vector_vector( R, dims );
            phat::chunk_reduction reduction_algorithm;
	    float size_b=boundary_matrix.get_num_entries();
	    float size_gig=size_b*8/(1024*1024);
	    std::cerr << "Size of boundary matrix : " << size_gig << " bytes" << std::endl;
            reduction_algorithm( boundary_matrix );
            boundary_matrix.save_vector_vector( R, dims );

            std::cerr << "done\n Time for reduction: " << overall_timer.elapsed() << std::endl;

            /*#if PRINT_MATRICES
            print_matrix(R);
            #endif*/
            for(Index j=0;j<R.size();j++) {
                if(!is_zero(R[j])) {

                    Index i = lowest(R[j]);

                    //std::cout << i << " " << j << std::endl;
                    //continue;

#if WITH_FLIPPING
                    Smplx b=v[R.size()-j-1];
                    Smplx d=v[R.size()-i-1];
#else
                    Smplx b=v[i];
                    Smplx d=v[j];
#endif
                    float birth = data(b);
                    float death = data(d);
                    bool birth_is_coned = is_coned(b);
                    bool death_is_coned = is_coned(d);
                    if(birth_is_coned==death_is_coned && birth==death) {
                        continue;
                    }
                    float curr_pers = (death-birth);

                    total_persistence[0]+=1.;
                    total_persistence[1]+=curr_pers;
                    total_persistence[2]+=curr_pers*curr_pers;
                    if(! birth_is_coned && ! death_is_coned) {
		      std::cout << "regular " << std::flush;
                        no_regular++;
                    } else if(!birth_is_coned && death_is_coned) {
		      std::cout << "extended " << std::flush;
                        no_extended++;
                    } else {
                        assert(birth_is_coned && death_is_coned);
			std::cout << "relative " << std::flush;
                        no_relative++;
                    }
                    no_dim[dim(b)]++;
		    std::cout << dim(b) << " " 
                        << std::setprecision (std::numeric_limits<double>::digits10 + 1) << data(b) << " " 
			      << data(d)<<"\n";/* << "\t"<< i <<"\t"<< j <<"\t"<<"<"<<std::flush;
			//Smplx::first_type::const_iterator it = b.first.begin(); std::cout<< *it << std::endl;
			for(Smplx::first_type::const_iterator it=b.first.begin(); it!=b.first.end(); ++it)
				{
						if(it==b.first.end()){std::cout<<*it<<">";}
						else{std::cout<<*it<<" ";}
				}
				std::cout<<">"<<"\t"<<"<"<<std::flush;
				//Smplx::first_type::const_iterator it1=b.first.end(); std::cout<<*it1<<">"<<"\t"<<"<"<<std::flush;
			for(Smplx::first_type::const_iterator it=d.first.begin(); it!=d.first.end(); ++it)
				{
						if(it==d.first.end()){std::cout<<*it<<">";}
						else{std::cout<<*it<<" ";}
				}
				std::cout<<">"<<std::endl;*/
				//it1=d.first.end(); std::cout<<*it1<<">"<<std::endl;	
                }
				

            }
	    std::cout << "end"<<std::endl;
            std::cerr << "INFO: #regular dots " << (no_regular) << std::endl;
            std::cerr << "INFO: #extended dots " << (no_extended) << std::endl;
            std::cerr << "INFO: #relative dots " << (no_relative) << std::endl;
            std::cerr << "INFO: #dim 0 dots " << (no_dim[0]) << std::endl;
            std::cerr << "INFO: #dim 1 dots " << (no_dim[1]) << std::endl;
            std::cerr << "INFO: #dim 2 dots " << (no_dim[2]) << std::endl;
            std::cerr << "INFO: #dim 3 dots " << (no_dim[3]) << std::endl;
            std::cerr << "INFO: #dots " << (no_dim[0]+no_dim[1]+no_dim[2]+no_dim[3]) << std::endl;
            std::cerr << "INFO: degree 0 total persistence " << (total_persistence[0]) << std::endl;
            std::cerr << "INFO: degree 1 total persistence " << (total_persistence[1]) << std::endl;
            std::cerr << "INFO: degree 2 total persistence " << (total_persistence[2]) << std::endl;
            std::cerr << "INFO: Time to compute persistence " << (overall_timer.elapsed()) << std::endl;
        }
        if(indicator=='s') {
            Smplx smplx;
            set_data(smplx,curr_val);
            unsigned s,s_old=-1;		
            while(sstr) {
                sstr >> s;
                if(s!=s_old) {
                    add_vertex(smplx,s);
                }
                s_old=s;
            }
            //Smplx smpl(vertices.begin(),vertices.end(),curr_val);
            v.push_back(smplx);
        }
    }

    //inFile.close();
} // end of the method handle  



int main(int argc, char** argv) {

  /*handle(argv[1]);
#if COUNT_ADDITIONS
    std::cerr << "Additions: " << add_basic << std::endl;
#endif
return 0;*/

  while(std::cin) {
      handle();
    }
    return 0;
}
